/*
 * unyaffs: extract files from yaffs2 file system image to current directory
 *
 * Created by Kai Wei <kai.wei.cn@gmail.com>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <utime.h>
#include "unyaffs.h"

#define CHUNK_SIZE 2048
#define SPARE_SIZE 64
#define MAX_OBJECTS 10000
#define YAFFS_OBJECTID_ROOT     1

typedef struct {
				char path[256];
				time_t atime;
				time_t mtime;
} path_entry;

path_entry paths[1024];
int path_count=0;

unsigned char data[CHUNK_SIZE + SPARE_SIZE];
unsigned char *chunk_data = data;
unsigned char *spare_data = data + CHUNK_SIZE;
int img_file;
int verbose = 0;
int dry_run = 0;
char perms_tpl[] = "rwxrwxrwx";
char perms_str[sizeof(perms_tpl)];

extern int read_chunk();

char *obj_list[MAX_OBJECTS];

void do_perms(int perms)
{
	int i,j,k;

	strcpy(perms_str, perms_tpl);
	i = 8;
	for(j=0; j<9; j++) {
		if((perms & 1) == 0) {
			perms_str[i] = '-';
		}
		i--;
		perms >>= 1;
	}
}


int process_chunk()
{
	int out_file, remain;
	ssize_t s;
	size_t written, to_write;
	char *full_path_name;
	struct timespec times[2];
	int ret;


	yaffs_PackedTags2 *pt = (yaffs_PackedTags2 *)spare_data;
	if (pt->t.byteCount == 0xffff)  {	//a new object 

		yaffs_ObjectHeader *oh = (yaffs_ObjectHeader *)malloc(sizeof(yaffs_ObjectHeader));
		memcpy(oh, chunk_data, sizeof(yaffs_ObjectHeader));

		full_path_name = (char *)malloc(strlen(oh->name) + strlen(obj_list[oh->parentObjectId]) + 2);
		if (full_path_name == NULL) {
			perror("malloc full path name\n");
		}
		strcpy(full_path_name, obj_list[oh->parentObjectId]);
		strcat(full_path_name, "/");
		strcat(full_path_name, oh->name);
		obj_list[pt->t.objectId] = full_path_name;

		times[0].tv_sec = oh->yst_atime;
		times[0].tv_nsec = 0;
		times[1].tv_sec = oh->yst_mtime;
		times[1].tv_nsec = 0;

		if (verbose) {
			do_perms(oh->yst_mode);
			printf("%s %4d %4d ", perms_str, oh->yst_uid, oh->yst_gid);
		}

		switch(oh->type) {
			case YAFFS_OBJECT_TYPE_FILE:
				if (verbose)
					printf("Creating file %s\n", full_path_name);

				remain = oh->fileSize;
				if (!dry_run)
					out_file = creat(full_path_name, oh->yst_mode);
				while(remain > 0) {
					if (read_chunk() == -1) {
						return -1;
					}
					to_write = (remain < pt->t.byteCount) ? remain : pt->t.byteCount;
					written = 0;
					while (written != to_write) { 
						if (!dry_run) {
							s = write(out_file, chunk_data + written, to_write - written);
						} else {
							s = to_write - written;
						}
						if (s == -1) {
							return -1;
						}
						written += s;
					}
					remain -= s;
				}
				if (!dry_run) {
					close(out_file);
					if(geteuid() == 0) {
						ret = fchownat(AT_FDCWD, full_path_name, oh->yst_uid, oh->yst_gid, 0);
					}
					utimensat(AT_FDCWD, full_path_name, times, 0);
				}

				break;
			case YAFFS_OBJECT_TYPE_SYMLINK:
				if (verbose)
					printf("Creating symlink %s\n", full_path_name);

				if (!dry_run) {
					ret = symlink(oh->alias, full_path_name);
					if(geteuid() == 0) {
						ret = fchownat(AT_FDCWD, full_path_name, oh->yst_uid, oh->yst_gid, AT_SYMLINK_NOFOLLOW);
					}
					utimensat(AT_FDCWD, full_path_name, times, AT_SYMLINK_NOFOLLOW);
				}
				break;
			case YAFFS_OBJECT_TYPE_DIRECTORY:
				if (verbose)
					printf("Creating directory %s\n", full_path_name);

				if (!dry_run) {
					mkdir(full_path_name, oh->yst_mode);
					if(geteuid() == 0) {
						ret = fchownat(AT_FDCWD, full_path_name, oh->yst_uid, oh->yst_gid, 0);
					}
					strcpy(paths[path_count].path, full_path_name);
					paths[path_count].atime = oh->yst_atime;
					paths[path_count].mtime = oh->yst_mtime;
					path_count++;

/*					utimensat(AT_FDCWD, full_path_name, times, 0);*/
				}

				break;
			case YAFFS_OBJECT_TYPE_HARDLINK:
				if (verbose)
					printf("Creating hardlink %s\n", full_path_name);
				if (!dry_run) {
					ret = link(obj_list[oh->equivalentObjectId], full_path_name);
					if(geteuid() == 0) {
						ret = fchownat(AT_FDCWD, full_path_name, oh->yst_uid, oh->yst_gid, 0);
					}
					utimensat(AT_FDCWD, full_path_name, times, 0);
				}
				break;
			case YAFFS_OBJECT_TYPE_UNKNOWN:
				if (verbose)
					printf("ignored YAFFS_OBJECT_TYPE_UNKNOWN\n");

				break;
			case YAFFS_OBJECT_TYPE_SPECIAL:
				if (verbose)
					printf("ignored YAFFS_OBJECT_TYPE_SPECIAL\n");

				break;
		}
	}
	return 0;
}


int read_chunk()
{
	ssize_t s;
  size_t bytes_read, to_read;

	memset(chunk_data, 0xff, sizeof(chunk_data));
	to_read = CHUNK_SIZE + SPARE_SIZE;
	bytes_read = 0;
	while (bytes_read != to_read) {
		s = read(img_file, data + bytes_read, CHUNK_SIZE + SPARE_SIZE);
		if (s == -1) {
			printf("failed in read image\n");
			perror("read image file\n");
	                return -1;
		}
		if (s == 0) {
			if (verbose)
				printf("end of image\n");
			return -1;
		}
		bytes_read += s;

	}
	return 0;
}

void usage(char **argv)
{
	printf("Usage: %s [OPTION]... [IMAGE]\n", argv[0]);
	printf("Extract files from a YAFFS IMAGE (now only support images created by mkyaffs2image)\n\n");
	printf("\t --dry_run\t do not extract the files. Typically used in conjuntion with --verbose\n");
	printf("\t --verbose\t be verbose about the extraction\n");
	printf("\t --help\t display this help and exit\n");
}

int main(int argc, char **argv)
{
	int options_idx = 0;
	struct timespec times[2];
	int i;

	while (++options_idx < argc) {
		if (strncmp( argv[options_idx], "-", 1) == 0)
		{
			if (strcmp( argv[options_idx], "-h") == 0 || strcmp( argv[options_idx], "--help") == 0)
			{
				usage(argv);
				exit(0);
			} else if (strcmp(argv[options_idx], "--verbose") == 0)
			{
				verbose = 1;
			} else if (strcmp(argv[options_idx], "--dry_run") == 0)
			{
				dry_run = 1;
			} else
			{
				printf("%s: unrecognized option '%s'\n", argv[0], argv[options_idx]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				exit(1);
			}
	        } else {
			break;
		}
	}

	if(argv[options_idx]) {
		img_file = open(argv[options_idx], O_RDONLY);
	}
	else {
		printf("No image file given. Aborting.\n");
		exit(1);
	}

	if (img_file == -1) {
		printf("open image file '%s' failed\n", argv[options_idx] );
		exit(1);
	}

	obj_list[YAFFS_OBJECTID_ROOT] = ".";
	while(read_chunk() != -1) {
		if (process_chunk() == -1)
			break;
	}
	close(img_file);

	while(--path_count > 0) {
    		/* puts(paths[path_count].path); */
		times[0].tv_sec = paths[path_count].atime;
		times[0].tv_nsec = 0;
		times[1].tv_sec = paths[path_count].mtime;
		times[1].tv_nsec = 0;
		utimensat(AT_FDCWD, paths[path_count].path, times, 0);
	}
	return 0;
}
